import datetime
import uuid
from dataclasses import dataclass
from xml.etree.ElementTree import Element, SubElement

from bs4 import BeautifulSoup

from opentakserver.functions import iso8601_string_from_datetime
from opentakserver.extensions import db, logger
from sqlalchemy import Integer, String, Boolean, ForeignKey, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.models.Mission import Mission
from opentakserver.models.MissionContent import MissionContent
from opentakserver.models.MissionUID import MissionUID


@dataclass
class MissionChange(db.Model):
    __tablename__ = "mission_changes"

    CREATE_MISSION = "CREATE_MISSION"
    DELETE_MISSION = "DELETE_MISSION"
    ADD_CONTENT = "ADD_CONTENT"
    REMOVE_CONTENT = "REMOVE_CONTENT"
    CREATE_DATA_FEED = "CREATE_DATA_FEED"
    DELETE_DATA_FEED = "DELETE_DATA_FEED"
    CHANGE = "CHANGE"

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    content_uid: Mapped[str] = mapped_column(String(255), ForeignKey("mission_content.uid"), nullable=True)
    isFederatedChange: Mapped[bool] = mapped_column(Boolean)
    change_type: Mapped[str] = mapped_column(String(255))
    mission_name: Mapped[str] = mapped_column(String(255), ForeignKey('missions.name'))
    timestamp: Mapped[datetime] = mapped_column(DateTime)
    creator_uid: Mapped[str] = mapped_column(String(255))
    server_time: Mapped[datetime] = mapped_column(DateTime)
    mission_uid: Mapped[str] = mapped_column(String(255), ForeignKey('mission_uids.uid', ondelete="CASCADE"), nullable=True)
    content_resource = relationship("MissionContent", back_populates="mission_changes", uselist=False)
    mission = relationship("Mission", back_populates="mission_changes")
    uid = relationship("MissionUID", back_populates="mission_change", uselist=False)

    def serialize(self):
        return {
            "isFederatedChange": self.isFederatedChange,
            "change_type": self.change_type,
            "mission_name": self.mission_name,
            "timestamp": self.timestamp,
            "creator_uid": self.creator_uid,
            "server_time": self.server_time,
            "mission_uid": self.mission_uid,
            "content_uid": self.content_uid
        }

    def to_json(self):
        json = {
            "isFederatedChange": self.isFederatedChange,
            "type": self.change_type,
            "contentUid": self.content_uid,
            "missionName": self.mission_name,
            "timestamp": iso8601_string_from_datetime(self.timestamp),
            "creatorUid": self.creator_uid if self.creator_uid else "",
            "serverTime": iso8601_string_from_datetime(self.server_time),
            "missionGuid": self.mission_uid
        }

        if self.content_resource:
            json['contentResource'] = self.content_resource.to_json()['data']

        if self.uid:
            json['details'] = self.uid.to_details_json()

        return json


def generate_mission_change_cot(author_uid: str, mission: Mission, mission_change: MissionChange,
                                content: MissionContent | None = None, cot_event: BeautifulSoup | None = None,
                                mission_uid: MissionUID = None, cot_type: str = "t-x-m-c") -> Element:
    if content:
        uid = content.uid
    elif cot_event:
        uid = cot_event.attrs['uid']
    else:
        uid = str(uuid.uuid4())

    event = Element("event", {"version": "2.0", "uid": uid, "type": cot_type, "how": "h-g-i-g-o",
                              "start": iso8601_string_from_datetime(mission_change.timestamp),
                              "time": iso8601_string_from_datetime(mission_change.timestamp),
                              "stale": iso8601_string_from_datetime(
                                  mission_change.timestamp + datetime.timedelta(minutes=2)),
                              "access": "Undefined"})
    SubElement(event, "point", {"ce": "9999999", "le": "9999999", "hae": "0.0", "lat": "0.0", "lon": "0.0"})

    detail = SubElement(event, "detail")
    mission_element = SubElement(detail, "mission",
                                 {"type": mission_change.change_type, "tool": "public", "name": mission.name,
                                  "guid": mission.guid, "authorUid": mission_change.creator_uid})
    mission_changes_element = SubElement(mission_element, "MissionChanges")
    mission_change_element = SubElement(mission_changes_element, "MissionChange")

    if content:
        content_resource = SubElement(mission_change_element, "contentResource")
        SubElement(content_resource, "creatorUid").text = mission_change.creator_uid
        SubElement(content_resource, "expiration").text = "-1"
        SubElement(content_resource, "groupVector").text = "0"
        SubElement(content_resource, "hash").text = content.hash
        SubElement(content_resource, "mimeType").text = content.mime_type
        SubElement(content_resource, "name").text = content.filename
        SubElement(content_resource, "size").text = str(content.size)
        SubElement(content_resource, "submissionTime").text = iso8601_string_from_datetime(content.submission_time)
        SubElement(content_resource, "submitter").text = content.submitter
        SubElement(content_resource, "uid").text = content.uid
        SubElement(mission_change_element, "contentUid").text = mission_change.content_uid

    if cot_event:
        details_tag = SubElement(mission_change_element, "details", {'type': cot_event.attrs['type']})

        point = cot_event.find("point")
        color = cot_event.find("color")
        callsign = cot_event.find("contact")
        icon = cot_event.find("usericon")

        if color and 'argb' in color.attrs:
            details_tag.set("color", color.attrs['argb'])
        if color and 'value' in color.attrs:
            details_tag.set("color", color.attrs['value'])
        if callsign:
            details_tag.set("callsign", callsign.attrs['callsign'])
        if icon:
            details_tag.set("iconsetPath", icon.attrs['iconsetpath'])

        SubElement(details_tag, "location", {'lon': point.attrs['lon'], 'lat': point.attrs['lat']})
        # SubElement(mission_change_element, "contentUid").text = cot_event.attrs['uid']

    if mission_uid:
        details_tag = SubElement(mission_change_element, "details")
        if mission_uid.color:
            details_tag.set('color', str(mission_uid.color))
        if mission_uid.callsign:
            details_tag.set('callsign', mission_uid.callsign)
        if mission_uid.cot_type:
            details_tag.set('type', mission_uid.cot_type)
        if mission_uid.iconset_path:
            details_tag.set('iconsetPath', mission_uid.iconset_path)
        if mission_uid.longitude:
            SubElement(details_tag, "location", {'lon': str(mission_uid.longitude), 'lat': str(mission_uid.latitude)})

    if mission_change.mission_uid:
        SubElement(mission_change_element, "contentUid").text = mission_change.mission_uid

    SubElement(mission_change_element, "missionGuid").text = mission.guid
    SubElement(mission_change_element, "creatorUid").text = mission_change.creator_uid
    SubElement(mission_change_element, "isFederatedChange").text = str(mission_change.isFederatedChange).lower()
    SubElement(mission_change_element, "missionName").text = mission.name
    SubElement(mission_change_element, "timestamp").text = iso8601_string_from_datetime(mission_change.timestamp)
    SubElement(mission_change_element, "type").text = mission_change.change_type

    return event



"""Changed phone_number column to BigInteger for MySQL compatibility

Revision ID: 7e9f5d2c193d
Revises: 8787888a028f
Create Date: 2024-10-25 17:37:42.637254

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '7e9f5d2c193d'
down_revision = '8787888a028f'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('euds', schema=None) as batch_op:
        batch_op.alter_column('callsign',
               existing_type=sa.String(length=255),
               nullable=True)
        batch_op.alter_column('phone_number',
               existing_type=sa.INTEGER(),
               type_=sa.BigInteger(),
               existing_nullable=True)
    with op.batch_alter_table('data_packages', schema=None) as batch_op:
        batch_op.alter_column('creator_uid',
              existing_type=sa.String(length=255),
              nullable=True)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('euds', schema=None) as batch_op:
        batch_op.alter_column('phone_number',
               existing_type=sa.BigInteger(),
               type_=sa.INTEGER(),
               existing_nullable=True)
        batch_op.alter_column('callsign',
               existing_type=sa.String(length=255),
               nullable=True)
    with op.batch_alter_table('data_packages', schema=None) as batch_op:
        batch_op.alter_column('creator_uid',
              existing_type=sa.String(length=255),
              nullable=False)

    # ### end Alembic commands ###



"""Update for group support

Revision ID: e8bd877719e2
Revises: 3a7ea7ead17b
Create Date: 2025-10-08 18:56:33.910860

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'e8bd877719e2'
down_revision = '3a7ea7ead17b'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('groups_users',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.Column('group_id', sa.Integer(), nullable=False),
    sa.Column('direction', sa.String(length=255), nullable=False),
    sa.Column('enabled', sa.Boolean(), nullable=False),
    sa.ForeignKeyConstraint(['group_id'], ['groups.id'], ),
    sa.ForeignKeyConstraint(['user_id'], ['user.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.drop_table('groups_euds')

    with op.batch_alter_table('groups', schema=None) as batch_op:
        batch_op.add_column(sa.Column('name', sa.String(length=255), nullable=False))
        batch_op.add_column(sa.Column('distinguishedName', sa.String(length=255), nullable=True))
        batch_op.add_column(sa.Column('type', sa.String(length=255), nullable=False))
        batch_op.add_column(sa.Column('description', sa.String(length=255), nullable=True))
        batch_op.drop_column('group_type')
        batch_op.drop_column('group_name')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('groups', schema=None) as batch_op:
        batch_op.add_column(sa.Column('group_name', sa.VARCHAR(length=255), autoincrement=False, nullable=False))
        batch_op.add_column(sa.Column('group_type', sa.VARCHAR(length=255), autoincrement=False, nullable=False))
        batch_op.drop_column('description')
        batch_op.drop_column('type')
        batch_op.drop_column('distinguishedName')
        batch_op.drop_column('name')

    op.create_table('groups_euds',
    sa.Column('id', sa.INTEGER(), autoincrement=True, nullable=False),
    sa.Column('group_id', sa.INTEGER(), autoincrement=False, nullable=False),
    sa.Column('direction', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.Column('active', sa.BOOLEAN(), autoincrement=False, nullable=False),
    sa.Column('description', sa.VARCHAR(length=255), autoincrement=False, nullable=True),
    sa.Column('eud_uid', sa.VARCHAR(length=255), autoincrement=False, nullable=False),
    sa.ForeignKeyConstraint(['eud_uid'], ['euds.uid'], name=op.f('group_eud'), ondelete='CASCADE'),
    sa.ForeignKeyConstraint(['group_id'], ['groups.id'], name=op.f('group')),
    sa.PrimaryKeyConstraint('id', name=op.f('groups_euds_pkey'))
    )
    op.drop_table('groups_users')
    # ### end Alembic commands ###



"""Added groups table

Revision ID: d4cc3d4afdb9
Revises: d1f5df78eace
Create Date: 2024-10-15 19:03:59.996651

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd4cc3d4afdb9'
down_revision = 'd1f5df78eace'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('groups',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('group_name', sa.String(length=255), nullable=False),
    sa.Column('direction', sa.String(length=255), nullable=False),
    sa.Column('created', sa.DateTime(), nullable=False),
    sa.Column('group_type', sa.String(length=255), nullable=False),
    sa.Column('bitpos', sa.Integer(), nullable=False),
    sa.Column('active', sa.Boolean(), nullable=False),
    sa.Column('description', sa.String(length=255), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    op.drop_table('groups')
    # ### end Alembic commands ###



from flask import Blueprint
from flask_security import auth_required
from opentakserver.extensions import db
from opentakserver.models.EUDStats import EUDStats
from opentakserver.blueprints.ots_api.api import search, paginate

eud_stats_blueprint = Blueprint('eud_stats_blueprint', __name__)


@eud_stats_blueprint.route('/api/eud_stats')
@auth_required()
def get_stats():

    query = db.session.query(EUDStats)
    query = search(query, EUDStats, 'eud_uid')
    # TODO: Implement date range search
    #query = search(query, EUDStats, 'from')
    #query = search(query, EUDStats, 'to')

    rows = db.session.execute(query.order_by(EUDStats.id.desc()).limit(50)).all()

    results = {'results': [], 'total_pages': 1, 'current_page': 1, 'per_page': len(rows)}

    for row in rows:
        results['results'].insert(0, row[0].to_json())

    return results



"""Fixes to support MariaDB

Revision ID: 8787888a028f
Revises: f6dfc571d31c
Create Date: 2024-10-25 15:44:27.370608

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '8787888a028f'
down_revision = 'f6dfc571d31c'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('cot', schema=None) as batch_op:
        batch_op.alter_column('xml',
               existing_type=sa.String(length=255),
               type_=sa.TEXT(),
               existing_nullable=False)

    with op.batch_alter_table('euds', schema=None) as batch_op:
        batch_op.alter_column('meshtastic_id',
               existing_type=sa.INTEGER(),
               type_=sa.BigInteger(),
               existing_nullable=True)

    with op.batch_alter_table('points', schema=None) as batch_op:
        batch_op.alter_column('device_uid',
              existing_type=sa.String(length=255),
              nullable=True)

    with op.batch_alter_table('missions', schema=None) as batch_op:
        batch_op.create_foreign_key('mission_creator', 'euds', ['creator_uid'], ['uid'])

    with op.batch_alter_table('groups', schema=None) as batch_op:
        batch_op.add_column(sa.Column('eud_uid', sa.String(length=255), nullable=False))
        batch_op.create_foreign_key('eud', 'euds', ['eud_uid'], ['uid'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('missions', schema=None) as batch_op:
        batch_op.drop_constraint('mission_creator', type_='foreignkey')

    with op.batch_alter_table('euds', schema=None) as batch_op:
        batch_op.alter_column('meshtastic_id',
               existing_type=sa.BigInteger(),
               type_=sa.INTEGER(),
               existing_nullable=True)

    with op.batch_alter_table('cot', schema=None) as batch_op:
        batch_op.alter_column('xml',
               existing_type=sa.TEXT(),
               type_=sa.String(length=255),
               existing_nullable=False)

    with op.batch_alter_table('groups', schema=None) as batch_op:
        batch_op.drop_constraint('eud', type_='foreignkey')
        batch_op.drop_column('eud_uid')

    with op.batch_alter_table('points', schema=None) as batch_op:
        batch_op.alter_column('device_uid',
              existing_type=sa.String(length=255),
              nullable=False)

    # ### end Alembic commands ###



"""Initial Migration

Revision ID: 4c7909c34d4e
Revises: 
Create Date: 2024-05-24 04:08:57.447120

"""
from alembic import op
import sqlalchemy as sa
import flask_security
import logging
logger = logging.getLogger('OpenTAKServer')
logger.disabled = False


# revision identifiers, used by Alembic.
revision = '4c7909c34d4e'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('chatrooms',
    sa.Column('id', sa.String(length=255), nullable=False),
    sa.Column('name', sa.String(length=255), nullable=False),
    sa.Column('group_owner', sa.String(length=255), nullable=True),
    sa.Column('parent', sa.String(length=255), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )

    op.create_table('icons',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('iconset_uid', sa.String(length=255), nullable=True),
    sa.Column('filename', sa.String(length=255), nullable=True),
    sa.Column('groupName', sa.String(length=255), nullable=True),
    sa.Column('type2525b', sa.String(length=255), nullable=True),
    sa.Column('useCnt', sa.Integer(), nullable=True),
    sa.Column('bitmap', sa.LargeBinary(), nullable=True),
    sa.Column('shadow', sa.LargeBinary(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )

    op.create_table('role',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=80), nullable=False),
    sa.Column('description', sa.String(length=255), nullable=True),
    sa.Column('permissions', flask_security.datastore.AsaList(), nullable=True),
    sa.Column('update_datetime', sa.DateTime(), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('name')
    )

    op.create_table('user',
    sa.Column('email', sa.String(length=255), nullable=True),
    sa.Column('fs_webauthn_user_handle', sa.String(length=64), nullable=True),
    sa.Column('mf_recovery_codes', flask_security.datastore.AsaList(), nullable=True),
    sa.Column('password', sa.String(length=255), nullable=True),
    sa.Column('us_phone_number', sa.String(length=128), nullable=True),
    sa.Column('username', sa.String(length=255), nullable=True),
    sa.Column('us_totp_secrets', sa.Text(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('active', sa.Boolean(), nullable=False),
    sa.Column('fs_uniquifier', sa.String(length=64), nullable=False),
    sa.Column('confirmed_at', sa.DateTime(), nullable=True),
    sa.Column('last_login_at', sa.DateTime(), nullable=True),
    sa.Column('current_login_at', sa.DateTime(), nullable=True),
    sa.Column('last_login_ip', sa.String(length=64), nullable=True),
    sa.Column('current_login_ip', sa.String(length=64), nullable=True),
    sa.Column('login_count', sa.Integer(), nullable=True),
    sa.Column('tf_primary_method', sa.String(length=64), nullable=True),
    sa.Column('tf_totp_secret', sa.String(length=255), nullable=True),
    sa.Column('tf_phone_number', sa.String(length=128), nullable=True),
    sa.Column('create_datetime', sa.DateTime(), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=False),
    sa.Column('update_datetime', sa.DateTime(), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('fs_uniquifier'),
    sa.UniqueConstraint('fs_webauthn_user_handle'),
    sa.UniqueConstraint('us_phone_number'),
    sa.UniqueConstraint('username')
    )

    op.create_table('roles_users',
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.Column('role_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['role_id'], ['role.id'], name="role"),
    sa.ForeignKeyConstraint(['user_id'], ['user.id'], name="user")
    )

    op.create_table('teams',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=255), nullable=False),
    sa.Column('chatroom_id', sa.String(length=255), nullable=True),
    sa.ForeignKeyConstraint(['chatroom_id'], ['chatrooms.id'], name="team_chatroom"),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('name')
    )

    op.create_table('web_authn',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('credential_id', sa.LargeBinary(length=1024), nullable=False),
    sa.Column('public_key', sa.LargeBinary(), nullable=False),
    sa.Column('sign_count', sa.Integer(), nullable=True),
    sa.Column('transports', flask_security.datastore.AsaList(), nullable=True),
    sa.Column('backup_state', sa.Boolean(), nullable=False),
    sa.Column('device_type', sa.String(length=64), nullable=False),
    sa.Column('extensions', sa.String(length=255), nullable=True),
    sa.Column('create_datetime', sa.DateTime(), server_default=sa.text('(CURRENT_TIMESTAMP)'), nullable=False),
    sa.Column('lastuse_datetime', sa.DateTime(), nullable=False),
    sa.Column('name', sa.String(length=64), nullable=False),
    sa.Column('usage', sa.String(length=64), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['user_id'], ['user.id'], name="web_authn_user", ondelete='CASCADE'),
    sa.PrimaryKeyConstraint('id')
    )
    with op.batch_alter_table('web_authn', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_web_authn_credential_id'), ['credential_id'], unique=True)

    op.create_table('euds',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('uid', sa.String(length=255), nullable=False),
    sa.Column('callsign', sa.String(length=255), unique=True),
    sa.Column('device', sa.String(length=255), nullable=True),
    sa.Column('os', sa.String(length=255), nullable=True),
    sa.Column('platform', sa.String(length=255), nullable=True),
    sa.Column('version', sa.String(length=255), nullable=True),
    sa.Column('phone_number', sa.Integer(), nullable=True),
    sa.Column('last_event_time', sa.DateTime(), nullable=True),
    sa.Column('last_status', sa.String(length=255), nullable=True),
    sa.Column('user_id', sa.Integer(), nullable=True),
    sa.Column('team_id', sa.Integer(), nullable=True),
    sa.Column('team_role', sa.String(length=255), nullable=True),
    sa.ForeignKeyConstraint(['team_id'], ['teams.id'], name="eud_team"),
    sa.ForeignKeyConstraint(['user_id'], ['user.id'], name="eud_user"),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('uid')
    )

    op.create_table('chatrooms_uids',
    sa.Column('chatroom_id', sa.String(length=255), nullable=False),
    sa.Column('uid', sa.String(length=255), nullable=False),
    sa.ForeignKeyConstraint(['chatroom_id'], ['chatrooms.id'], name="chatroom_uids"),
    sa.ForeignKeyConstraint(['uid'], ['euds.uid'], name="chatroom_eud"),
    sa.PrimaryKeyConstraint('chatroom_id', 'uid')
    )

    op.create_table('cot',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('how', sa.String(length=255), nullable=True),
    sa.Column('type', sa.String(length=255), nullable=True),
    sa.Column('sender_callsign', sa.String(length=255), nullable=False),
    sa.Column('sender_device_name', sa.String(length=255), nullable=True),
    sa.Column('sender_uid', sa.String(length=255), nullable=True),
    sa.Column('recipients', sa.JSON(), nullable=True),
    sa.Column('timestamp', sa.DateTime(), nullable=False),
    sa.Column('start', sa.DateTime(), nullable=False),
    sa.Column('stale', sa.DateTime(), nullable=False),
    sa.Column('xml', sa.TEXT, nullable=False),
    sa.ForeignKeyConstraint(['sender_uid'], ['euds.uid'], name="cot_eud"),
    sa.PrimaryKeyConstraint('id')
    )

    op.create_table('data_packages',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('filename', sa.String(length=255), nullable=False),
    sa.Column('hash', sa.String(length=255), nullable=False),
    sa.Column('creator_uid', sa.String(length=255), nullable=False),
    sa.Column('submission_time', sa.DateTime(), nullable=False),
    sa.Column('submission_user', sa.Integer(), nullable=True),
    sa.Column('keywords', sa.String(length=255), nullable=True),
    sa.Column('mime_type', sa.String(length=255), nullable=False),
    sa.Column('size', sa.Integer(), nullable=False),
    sa.Column('tool', sa.String(length=255), nullable=True),
    sa.Column('expiration', sa.String(length=255), nullable=True),
    sa.ForeignKeyConstraint(['creator_uid'], ['euds.uid'], name="data_packages_eud"),
    sa.ForeignKeyConstraint(['submission_user'], ['user.id'], name="data_packages_user"),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('filename'),
    sa.UniqueConstraint('hash')
    )

    op.create_table('certificates',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('common_name', sa.String(length=255), nullable=False),
    sa.Column('eud_uid', sa.String(length=255), nullable=True),
    sa.Column('data_package_id', sa.Integer(), nullable=True),
    sa.Column('callsign', sa.String(length=255), nullable=True),
    sa.Column('username', sa.String(length=255), nullable=True),
    sa.Column('expiration_date', sa.DateTime(), nullable=False),
    sa.Column('server_address', sa.String(length=255), nullable=False),
    sa.Column('server_port', sa.Integer(), nullable=False),
    sa.Column('truststore_filename', sa.String(length=255), nullable=False),
    sa.Column('user_cert_filename', sa.String(length=255), nullable=False),
    sa.Column('csr', sa.String(length=255), nullable=True),
    sa.Column('cert_password', sa.String(length=255), nullable=False),
    sa.ForeignKeyConstraint(['data_package_id'], ['data_packages.id'], name="certificate_data_package"),
    sa.ForeignKeyConstraint(['eud_uid'], ['euds.uid'], name="certificate_eud"),
    sa.ForeignKeyConstraint(['username'], ['user.username'], name="certificate_user"),
    sa.PrimaryKeyConstraint('id')
    )

    op.create_table('points',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('uid', sa.String(length=255), nullable=False),
    sa.Column('device_uid', sa.String(length=255), nullable=False),
    sa.Column('latitude', sa.Float(), nullable=True),
    sa.Column('longitude', sa.Float(), nullable=True),
    sa.Column('ce', sa.Float(), nullable=True),
    sa.Column('hae', sa.Float(), nullable=True),
    sa.Column('le', sa.Float(), nullable=True),
    sa.Column('course', sa.Float(), nullable=True),
    sa.Column('speed', sa.Float(), nullable=True),
    sa.Column('location_source', sa.String(length=255), nullable=True),
    sa.Column('battery', sa.Float(), nullable=True),
    sa.Column('timestamp', sa.DateTime(), nullable=False),
    sa.Column('azimuth', sa.Float(), nullable=True),
    sa.Column('fov', sa.Float(), nullable=True),
    sa.Column('cot_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['cot_id'], ['cot.id'], name="point_cot"),
    sa.ForeignKeyConstraint(['device_uid'], ['euds.uid'], name="point_eud"),
    sa.PrimaryKeyConstraint('id')
    )

    op.create_table('video_streams',
    sa.Column('path', sa.String(length=255), nullable=False),
    sa.Column('protocol', sa.String(length=255), nullable=False),
    sa.Column('port', sa.Integer(), nullable=False),
    sa.Column('network_timeout', sa.Integer(), nullable=False),
    sa.Column('uid', sa.String(length=255), nullable=True),
    sa.Column('buffer_time', sa.Integer(), nullable=True),
    sa.Column('rover_port', sa.Integer(), nullable=True),
    sa.Column('rtsp_reliable', sa.Integer(), nullable=True),
    sa.Column('ignore_embedded_klv', sa.Boolean(), nullable=True),
    sa.Column('alias', sa.String(length=255), nullable=True),
    sa.Column('preferred_mac_address', sa.String(length=255), nullable=True),
    sa.Column('preferred_interface_address', sa.String(length=255), nullable=True),
    sa.Column('username', sa.String(length=255), nullable=True),
    sa.Column('xml', sa.String(length=255), nullable=True),
    sa.Column('ready', sa.Boolean(), nullable=False),
    sa.Column('mediamtx_settings', sa.String(length=255), nullable=False),
    sa.Column('cot_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['cot_id'], ['cot.id'], name="video_stream_cot"),
    sa.ForeignKeyConstraint(['username'], ['user.username'], name="video_stream_user"),
    sa.PrimaryKeyConstraint('path')
    )

    op.create_table('alerts',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('uid', sa.String(length=255), nullable=False),
    sa.Column('sender_uid', sa.String(length=255), nullable=False),
    sa.Column('start_time', sa.DateTime(), nullable=False),
    sa.Column('cancel_time', sa.DateTime(), nullable=True),
    sa.Column('alert_type', sa.String(length=255), nullable=False),
    sa.Column('point_id', sa.Integer(), nullable=True),
    sa.Column('cot_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['cot_id'], ['cot.id'], name="alert_cot"),
    sa.ForeignKeyConstraint(['point_id'], ['points.id'], name="alert_point"),
    sa.ForeignKeyConstraint(['sender_uid'], ['euds.uid'], name="alert_eud"),
    sa.PrimaryKeyConstraint('id')
    )

    op.create_table('casevac',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('sender_uid', sa.String(length=255), nullable=False),
    sa.Column('uid', sa.String(length=255), nullable=False),
    sa.Column('timestamp', sa.DateTime(), nullable=False),
    sa.Column('ambulatory', sa.Integer(), nullable=True),
    sa.Column('casevac', sa.Boolean(), nullable=True),
    sa.Column('child', sa.Integer(), nullable=True),
    sa.Column('enemy', sa.String(length=255), nullable=True),
    sa.Column('epw', sa.Integer(), nullable=True),
    sa.Column('equipment_detail', sa.String(length=255), nullable=True),
    sa.Column('equipment_none', sa.Boolean(), nullable=True),
    sa.Column('equipment_other', sa.Boolean(), nullable=True),
    sa.Column('extraction_equipment', sa.Boolean(), nullable=True),
    sa.Column('freq', sa.Float(), nullable=True),
    sa.Column('friendlies', sa.String(length=255), nullable=True),
    sa.Column('hlz_marking', sa.Integer(), nullable=True),
    sa.Column('hlz_remarks', sa.String(length=255), nullable=True),
    sa.Column('hoist', sa.Boolean(), nullable=True),
    sa.Column('litter', sa.Integer(), nullable=True),
    sa.Column('marked_by', sa.String(length=255), nullable=True),
    sa.Column('medline_remarks', sa.String(length=255), nullable=True),
    sa.Column('nonus_civilian', sa.Integer(), nullable=True),
    sa.Column('nonus_military', sa.Integer(), nullable=True),
    sa.Column('obstacles', sa.String(length=255), nullable=True),
    sa.Column('priority', sa.Integer(), nullable=True),
    sa.Column('routine', sa.Integer(), nullable=True),
    sa.Column('security', sa.Integer(), nullable=True),
    sa.Column('terrain_loose', sa.Boolean(), nullable=True),
    sa.Column('terrain_other', sa.Boolean(), nullable=True),
    sa.Column('terrain_other_detail', sa.String(length=255), nullable=True),
    sa.Column('terrain_detail', sa.String(length=255), nullable=True),
    sa.Column('terrain_none', sa.Boolean(), nullable=True),
    sa.Column('terrain_rough', sa.Boolean(), nullable=True),
    sa.Column('terrain_slope', sa.Boolean(), nullable=True),
    sa.Column('terrain_slope_dir', sa.String(length=255), nullable=True),
    sa.Column('title', sa.String(length=255), nullable=False),
    sa.Column('urgent', sa.Integer(), nullable=True),
    sa.Column('us_civilian', sa.Integer(), nullable=True),
    sa.Column('us_military', sa.Integer(), nullable=True),
    sa.Column('ventilator', sa.Boolean(), nullable=True),
    sa.Column('winds_are_from', sa.String(length=255), nullable=True),
    sa.Column('zone_prot_selection', sa.Integer(), nullable=True),
    sa.Column('point_id', sa.Integer(), nullable=True),
    sa.Column('cot_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['cot_id'], ['cot.id'], name="casevac_cot"),
    sa.ForeignKeyConstraint(['point_id'], ['points.id'], name="casevac_point"),
    sa.ForeignKeyConstraint(['sender_uid'], ['euds.uid'], name="casevac_eud"),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('uid')
    )

    op.create_table('geochat',
    sa.Column('uid', sa.String(length=255), nullable=False),
    sa.Column('chatroom_id', sa.String(length=255), nullable=False),
    sa.Column('sender_uid', sa.String(length=255), nullable=False),
    sa.Column('remarks', sa.String(length=255), nullable=False),
    sa.Column('timestamp', sa.DateTime(), nullable=False),
    sa.Column('point_id', sa.Integer(), nullable=False),
    sa.Column('cot_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['chatroom_id'], ['chatrooms.id'], name="geochat_chatroom"),
    sa.ForeignKeyConstraint(['cot_id'], ['cot.id'], name="geochat_cot"),
    sa.ForeignKeyConstraint(['point_id'], ['points.id'], name="geochat_point"),
    sa.ForeignKeyConstraint(['sender_uid'], ['euds.uid'], name="geochat_eud"),
    sa.PrimaryKeyConstraint('uid')
    )

    op.create_table('markers',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('uid', sa.String(length=255), nullable=False),
    sa.Column('affiliation', sa.String(length=255), nullable=True),
    sa.Column('battle_dimension', sa.String(length=255), nullable=True),
    sa.Column('point_id', sa.Integer(), nullable=False),
    sa.Column('callsign', sa.String(length=255), nullable=True),
    sa.Column('readiness', sa.Boolean(), nullable=True),
    sa.Column('argb', sa.Integer(), nullable=True),
    sa.Column('color_hex', sa.String(length=255), nullable=True),
    sa.Column('iconset_path', sa.String(length=255), nullable=True),
    sa.Column('parent_callsign', sa.String(length=255), nullable=True),
    sa.Column('production_time', sa.String(length=255), nullable=True),
    sa.Column('relation', sa.String(length=255), nullable=True),
    sa.Column('relation_type', sa.String(length=255), nullable=True),
    sa.Column('location_source', sa.String(length=255), nullable=True),
    sa.Column('icon_id', sa.Integer(), nullable=True),
    sa.Column('parent_uid', sa.String(length=255), nullable=True),
    sa.Column('remarks', sa.String(length=255), nullable=True),
    sa.Column('cot_id', sa.Integer(), nullable=True),
    sa.Column('mil_std_2525c', sa.String(length=255), nullable=True),
    sa.ForeignKeyConstraint(['cot_id'], ['cot.id'], name="marker_cot"),
    sa.ForeignKeyConstraint(['icon_id'], ['icons.id'], name="marker_icon"),
    sa.ForeignKeyConstraint(['parent_uid'], ['euds.uid'], name="marker_eud"),
    sa.ForeignKeyConstraint(['point_id'], ['points.id'], name="marker_point"),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('uid')
    )

    op.create_table('rb_lines',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('sender_uid', sa.String(length=255), nullable=False),
    sa.Column('uid', sa.String(length=255), nullable=False),
    sa.Column('timestamp', sa.DateTime(), nullable=False),
    sa.Column('range', sa.Float(), nullable=False),
    sa.Column('bearing', sa.Float(), nullable=False),
    sa.Column('inclination', sa.Float(), nullable=True),
    sa.Column('anchor_uid', sa.String(length=255), nullable=True),
    sa.Column('range_units', sa.Integer(), nullable=True),
    sa.Column('bearing_units', sa.Integer(), nullable=True),
    sa.Column('north_ref', sa.Integer(), nullable=True),
    sa.Column('color', sa.Integer(), nullable=True),
    sa.Column('color_hex', sa.String(length=255), nullable=True),
    sa.Column('callsign', sa.String(length=255), nullable=True),
    sa.Column('stroke_color', sa.Integer(), nullable=True),
    sa.Column('stroke_weight', sa.Float(), nullable=True),
    sa.Column('stroke_style', sa.String(length=255), nullable=True),
    sa.Column('labels_on', sa.Boolean(), nullable=True),
    sa.Column('point_id', sa.Integer(), nullable=True),
    sa.Column('cot_id', sa.Integer(), nullable=True),
    sa.Column('end_latitude', sa.Float(), nullable=True),
    sa.Column('end_longitude', sa.Float(), nullable=True),
    sa.ForeignKeyConstraint(['cot_id'], ['cot.id'], name="rb_line_cot"),
    sa.ForeignKeyConstraint(['point_id'], ['points.id'], name="rb_line_point"),
    sa.ForeignKeyConstraint(['sender_uid'], ['euds.uid'], name="rb_line_eud"),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('uid')
    )

    op.create_table('video_recordings',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('segment_path', sa.String(length=255), nullable=False),
    sa.Column('path', sa.String(length=255), nullable=False),
    sa.Column('in_progress', sa.Boolean(), nullable=False),
    sa.Column('start_time', sa.DateTime(), nullable=True),
    sa.Column('stop_time', sa.DateTime(), nullable=True),
    sa.Column('duration', sa.Integer(), nullable=True),
    sa.Column('width', sa.Integer(), nullable=True),
    sa.Column('height', sa.Integer(), nullable=True),
    sa.Column('video_codec', sa.String(length=255), nullable=True),
    sa.Column('video_bitrate', sa.Integer(), nullable=True),
    sa.Column('audio_codec', sa.String(length=255), nullable=True),
    sa.Column('audio_bitrate', sa.Integer(), nullable=True),
    sa.Column('audio_samplerate', sa.Integer(), nullable=True),
    sa.Column('audio_channels', sa.Integer(), nullable=True),
    sa.Column('file_size', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['path'], ['video_streams.path'], name="video_recording_video_stream"),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('segment_path')
    )

    op.create_table('zmist',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('i', sa.String(length=255), nullable=True),
    sa.Column('m', sa.String(length=255), nullable=True),
    sa.Column('s', sa.String(length=255), nullable=True),
    sa.Column('t', sa.String(length=255), nullable=True),
    sa.Column('title', sa.String(length=255), nullable=True),
    sa.Column('z', sa.Integer(), nullable=True),
    sa.Column('casevac_uid', sa.String(length=255), nullable=False),
    sa.ForeignKeyConstraint(['casevac_uid'], ['casevac.uid'], name="zmist_casevac"),
    sa.PrimaryKeyConstraint('id')
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('zmist')
    op.drop_table('video_recordings')
    op.drop_table('rb_lines')
    op.drop_table('markers')
    op.drop_table('geochat')
    op.drop_table('casevac')
    op.drop_table('alerts')
    op.drop_table('video_streams')
    op.drop_table('points')
    op.drop_table('certificates')
    op.drop_table('data_packages')
    op.drop_table('cot')
    op.drop_table('chatrooms_uids')
    op.drop_table('euds')
    with op.batch_alter_table('web_authn', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_web_authn_credential_id'))

    op.drop_table('web_authn')
    op.drop_table('teams')
    op.drop_table('roles_users')
    op.drop_table('user')
    op.drop_table('role')
    op.drop_table('icons')
    op.drop_table('chatrooms')
    # ### end Alembic commands ###



from dataclasses import dataclass

from opentakserver.extensions import db
from sqlalchemy import Integer, ForeignKey, String, Boolean
from sqlalchemy.orm import Mapped, mapped_column, relationship


@dataclass
class GroupUser(db.Model):
    __tablename__ = "groups_users"

    user_id: Mapped[int] = mapped_column(Integer, ForeignKey("user.id"), primary_key=True)
    group_id: Mapped[Integer] = mapped_column(Integer, ForeignKey("groups.id"), primary_key=True)
    direction: Mapped[String] = mapped_column(String(255), primary_key=True)
    enabled: Mapped[Boolean] = mapped_column(Boolean, default=True)
    user = relationship("User", cascade="all, delete", viewonly=True)
    group = relationship("Group", cascade="all, delete", viewonly=True)

    def serialize(self):
        return {
            "user": self.user.serialize(),
            "group": self.group.serialize(),
            "direction": self.direction,
            "active": self.enabled
        }

    def to_marti_json(self):
        return {
            "name": self.group.name,
            "direction": self.direction,
            "created": int(self.group.created.timestamp()),
            "type": self.group.type,
            "bitpos": self.group.bitpos,
            "active": self.enabled
        }

    def to_json(self):
        return {
            "username": self.user.username,
            "group_name": self.group.name,
            "direction": self.direction,
            "active": self.enabled
        }



import datetime
import hashlib
import json
import os
import platform
import traceback
from shutil import copyfile

from urllib.parse import urlparse

import pika
import yaml

import bleach
import psutil
import sqlalchemy.exc
from flask import current_app as app, request, Blueprint, jsonify, send_from_directory, session
from flask_ldap3_login import AuthenticationResponseStatus
from flask_security import auth_required, current_user, verify_password
from flask_babel import gettext
from sqlalchemy import select

from opentakserver.extensions import logger, db, ldap_manager, babel

from opentakserver.models.Alert import Alert
from opentakserver.models.CasEvac import CasEvac
from opentakserver.models.CoT import CoT
from opentakserver.models.DataPackage import DataPackage
from opentakserver.models.EUD import EUD
from opentakserver.models.Group import Group
from opentakserver.models.GroupUser import GroupUser
from opentakserver.models.Token import Token
from opentakserver.models.ZMIST import ZMIST
from opentakserver.models.Point import Point
from opentakserver.models.user import User
from opentakserver.models.Certificate import Certificate
from opentakserver.models.APSchedulerJobs import APSchedulerJobs

from opentakserver.certificate_authority import CertificateAuthority
from opentakserver.models.Icon import Icon
from opentakserver.models.Marker import Marker
from opentakserver.models.RBLine import RBLine
from opentakserver import __version__ as version

api_blueprint = Blueprint('api_blueprint', __name__)

p = psutil.Process()


def search(query, model, field):
    arg = request.args.get(field)
    if arg:
        arg = bleach.clean(arg)
        return query.where(getattr(model, field) == arg)
    return query


def paginate(query):
    try:
        page = int(request.args.get('page')) if 'page' in request.args else 1
        per_page = int(request.args.get('per_page')) if 'per_page' in request.args else 10
    except ValueError:
        return {'success': False, 'error': 'Invalid page or per_page number'}, 400, {'Content-Type': 'application/json'}

    pagination = db.paginate(query, page=page, per_page=per_page)
    rows = pagination.items

    results = {'results': [], 'total_pages': pagination.pages, 'current_page': page, 'per_page': per_page}

    for row in rows:
        results['results'].append(row.to_json())

    return jsonify(results)


def change_config_setting(setting, value):
    try:
        with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), "r") as config_file:
            config = yaml.safe_load(config_file.read())

        config[setting] = value
        with open(os.path.join(app.config.get("OTS_DATA_FOLDER"), "config.yml"), "w") as config_file:
            yaml.safe_dump(config, config_file)

    except BaseException as e:
        logger.error("Failed to change setting {} to {} in config.yml: {}".format(setting, value, e))


def route_cot(event: str, user: User):
    """Used by API endpoints such as ``/api/markers`` to route CoT messages to the correct groups that a user belongs to.

    :param event: The CoT to be routed as a string
    :param user: The user object
    :return: None
    """

    rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
    channel = rabbit_connection.channel()

    group_memberships = db.session.execute(db.session.query(GroupUser).filter_by(user_id=user.id, direction=Group.IN, enabled=True)).all()
    if not group_memberships:
        # Default to the __ANON__ group if the user doesn't belong to any IN groups
        channel.basic_publish(exchange='groups', routing_key="__ANON__.OUT", body=json.dumps({'uid': app.config['OTS_NODE_ID'], 'cot': str(event)}),
                              properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")))

    for membership in group_memberships:
        membership = membership[0]
        channel.basic_publish(exchange='groups', routing_key=f"{membership.group.name}.{Group.OUT}",
                              body=json.dumps({'uid': app.config['OTS_NODE_ID'], 'cot': str(event)}),
                              properties=pika.BasicProperties(expiration=app.config.get("OTS_RABBITMQ_TTL")))
    channel.close()
    rabbit_connection.close()


@api_blueprint.route('/files/api/config')
def cloudtak_config():
    """Required by CloudTAK, returns the following

    .. code-block:: json

        {"uploadSizeLimit": 400}
    """
    return jsonify({'uploadSizeLimit': 400})


# Simple health check for docker
@api_blueprint.route('/api/health')
def health():
    """Health check for Docker, returns the following

    .. code-block:: json
        {"status": "healthy"}
    """
    return jsonify({'status': 'healthy'})


@api_blueprint.route('/api/status')
@auth_required()
def status():
    """Server status used on the Dashboard page of the web UI

    :rtype: dict
    """
    now = datetime.datetime.now(datetime.timezone.utc)
    system_boot_time = datetime.datetime.fromtimestamp(psutil.boot_time(), datetime.datetime.now().astimezone().tzinfo)
    system_uptime = now - system_boot_time

    ots_uptime = now - app.start_time

    cpu_time = psutil.cpu_times()
    cpu_time_dict = {'user': cpu_time.user, 'system': cpu_time.system, 'idle': cpu_time.idle}

    vmem = psutil.virtual_memory()
    vmem_dict = {'total': vmem.total, 'available': vmem.available, 'used': vmem.used, 'free': vmem.free,
                 'percent': vmem.percent}

    disk_usage = psutil.disk_usage('/')
    disk_usage_dict = {'total': disk_usage.total, 'used': disk_usage.used, 'free': disk_usage.free,
                       'percent': disk_usage.percent}

    try:
        os_release = platform.freedesktop_os_release()
    except:
        os_release = {'NAME': None, 'PRETTY_NAME': None, 'VERSION': None, 'VERSION_CODENAME': None}

    uname = {'system': platform.system(), 'node': platform.node(), 'release': platform.release(),
             'version': platform.version(), 'machine': platform.machine()}

    online_euds = db.session.execute(select(EUD).filter(EUD.last_status == 'Connected')).all()

    response = {
        'online_euds': len(online_euds), 'system_boot_time': system_boot_time.strftime("%Y-%m-%d %H:%M:%SZ"),
        'system_uptime': system_uptime.total_seconds(), 'ots_start_time': app.start_time.strftime("%Y-%m-%d %H:%M:%SZ"),
        'ots_uptime': ots_uptime.total_seconds(), 'cpu_time': cpu_time_dict, 'cpu_percent': p.cpu_percent(),
        'load_avg': psutil.getloadavg(), 'memory': vmem_dict, 'disk_usage': disk_usage_dict, 'ots_version': version,
        'uname': uname, 'os_release': os_release, 'python_version': platform.python_version()
    }

    return jsonify(response)


@api_blueprint.route("/api/certificate", methods=['GET', 'POST'])
@auth_required()
def certificate():
    if request.method == 'POST' and 'username' in request.json.keys():
        try:
            username = bleach.clean(request.json.get('username'))
            truststore_filename = os.path.join(app.config.get("OTS_CA_FOLDER"), 'certs', "opentakserver",
                                               "truststore-root.p12")
            user_filename = os.path.join(app.config.get("OTS_CA_FOLDER"), 'certs', username,
                                         "{}.p12".format(username))

            user = app.security.datastore.find_user(username=username)

            if not user:
                return jsonify({'success': False, 'error': gettext(u'Invalid username: %(username)s', username=username)}), 400

            ca = CertificateAuthority(logger, app)
            filenames = ca.issue_certificate(username, False)

            for filename in filenames:
                file_hash = hashlib.sha256(
                    open(os.path.join(app.config.get("OTS_CA_FOLDER"), 'certs', username, filename),
                         'rb').read()).hexdigest()

                data_package = DataPackage()
                data_package.filename = filename
                data_package.keywords = "public"
                data_package.creator_uid = request.json['uid'] if 'uid' in request.json.keys() else None
                data_package.submission_time = datetime.datetime.now(datetime.timezone.utc)
                data_package.mime_type = "application/x-zip-compressed"
                data_package.size = os.path.getsize(
                    os.path.join(app.config.get("OTS_CA_FOLDER"), 'certs', username, filename))
                data_package.hash = file_hash
                data_package.submission_user = current_user.id

                try:
                    db.session.add(data_package)
                    db.session.commit()
                except sqlalchemy.exc.IntegrityError as e:
                    db.session.rollback()
                    logger.error(e)
                    return jsonify({'success': False, 'error': gettext(u'Certificate already exists for %(username)s', username=username)}), 400

                copyfile(os.path.join(app.config.get("OTS_CA_FOLDER"), 'certs', username, "{}".format(filename)),
                         os.path.join(app.config.get("UPLOAD_FOLDER"), "{}.zip".format(file_hash)))

                cert = Certificate()
                cert.common_name = username
                cert.username = username
                cert.expiration_date = datetime.datetime.today() + datetime.timedelta(
                    days=app.config.get("OTS_CA_EXPIRATION_TIME"))
                cert.server_address = urlparse(request.url_root).hostname
                cert.server_port = app.config.get("OTS_SSL_STREAMING_PORT")
                cert.truststore_filename = truststore_filename
                cert.user_cert_filename = user_filename
                cert.cert_password = app.config.get("OTS_CA_PASSWORD")
                cert.data_package_id = data_package.id if data_package else None

                db.session.add(cert)
                db.session.commit()

            return jsonify({'success': True}), 200
        except BaseException as e:
            logger.error(traceback.format_exc())
            return jsonify({'success': False, 'error': str(e)}), 500
    elif request.method == 'POST':
        return jsonify({'success': False, 'error': gettext(u'Please specify a callsign')}), 400
    elif request.method == 'GET':
        query = db.session.query(Certificate)
        query = search(query, Certificate, 'callsign')
        query = search(query, Certificate, 'username')

        return paginate(query)


@api_blueprint.route('/api/me')
@auth_required()
def me():
    """Get the details of the currently logged in user

    :rtype: User
    """
    return jsonify(current_user.to_json())


@api_blueprint.route('/api/cot', methods=['GET'])
@auth_required()
def query_cot():
    """Get CoT messages. All parameters are optional.

    :param how: The how attribute of the CoT message, i.e. ``m-g``, ``h-g-i-g-o``
    :param type: The type attribute of the CoT message, i.e. ``a-f-G-U-C``
    :param sender_callsign: The callsign of the EUD that sent the CoT message
    :param sender_uid: The UID of the EUD that sent the CoT message
    :param page: The page number
    :param per_page: The number of results per page
    """
    query = db.session.query(CoT)
    query = search(query, CoT, 'how')
    query = search(query, CoT, 'type')
    query = search(query, CoT, 'sender_callsign')
    query = search(query, CoT, 'sender_uid')

    return paginate(query)


@api_blueprint.route("/api/alerts", methods=['GET'])
@auth_required()
def query_alerts():
    """Get alerts. All parameters are optional.

    :param uid: The alert's UID
    :param sender_uid: The UID of the EUD that sent the alert
    :param alert_type: The type of alert
    :param page: The page number
    :param per_page: The number of results per page
    """
    query = db.session.query(Alert)
    query = search(query, Alert, 'uid')
    query = search(query, Alert, 'sender_uid')
    query = search(query, Alert, 'alert_type')

    return paginate(query)


@api_blueprint.route("/api/point", methods=['GET'])
@auth_required()
def query_points():
    """Query points. All parameters are optional.

    :param uid: The point's UID
    :param callsign: The point's callsign
    :param page: The page number
    :param per_page: The number of results per page
    """
    query = db.session.query(Point)

    query = search(query, EUD, 'uid')
    query = search(query, EUD, 'callsign')

    return paginate(query)


@api_blueprint.route('/api/rabbitmq/<path>', methods=['POST'])
def rabbitmq_auth(path):
    # https://github.com/rabbitmq/rabbitmq-server/tree/v3.13.x/deps/rabbitmq_auth_backend_http

    # Only allow requests to this route from the RabbitMQ server
    if request.remote_addr != app.config.get("OTS_RABBITMQ_SERVER_ADDRESS"):
        return 'deny', 200

    username = bleach.clean(request.form.get('username'))
    password = None
    if 'password' in request.form.keys():
        password = bleach.clean(request.form.get('password'))

    if app.config.get("OTS_ENABLE_LDAP"):
        result = ldap_manager.authenticate(username, password)

        if result.status == AuthenticationResponseStatus.success:
            # Keep this import here to avoid a circular import when OTS is started
            from opentakserver.blueprints.ots_api.ldap_api import save_user

            save_user(result.user_dn, result.user_id, result.user_info, result.user_groups)

            for group in result.user_groups:
                if group['cn'] == app.config.get("OTS_LDAP_ADMIN_GROUP"):
                    return 'allow administrator', 200

            return 'allow', 200
        else:
            return 'deny', 200

    user = None
    if 'username' in request.form.keys():
        user = app.security.datastore.find_user(username=username)

    if user and 'password' in request.form.keys():
        if user.active and verify_password(password, user.password):
            if user.has_role("administrator"):
                return 'allow administrator', 200
            return 'allow', 200
        else:
            return 'deny', 200
    # Always allow when path is topic, resource, or vhost if the user exists.
    # This only occurs after the user has been successfully authenticated
    elif user:
        return 'allow', 200
    else:
        return 'deny', 200


@api_blueprint.route('/api/eud')
@auth_required()
def get_euds():
    """Query EUDS. All parameters are optional.

    :param callsign: The EUD's callsign
    :param uid: The EUD's callsign
    :param username: The username that the EUD belongs to
    :param page: The page number
    :param per_page: The number of results per page
    """
    query = db.session.query(EUD)

    if 'username' in request.args.keys():
        query = query.join(User, User.id == EUD.user_id)

    query = search(query, EUD, 'callsign')
    query = search(query, EUD, 'uid')
    query = search(query, User, 'username')

    return paginate(query)


@api_blueprint.route('/api/truststore')
def get_truststore():
    """Downloads the server's truststore with no authentication required."""
    filename = f"truststore_root_{urlparse(request.url_root).hostname}.p12"
    return send_from_directory(app.config.get("OTS_CA_FOLDER"), 'truststore-root.p12', download_name=filename,
                               as_attachment=True)


@api_blueprint.route('/api/map_state')
@auth_required()
def get_map_state():
    """Gets the latest data to be displayed on the web UI's map"""
    try:
        results = {'euds': [], 'markers': [], 'rb_lines': [], 'casevacs': []}

        euds = db.session.execute(db.session.query(EUD)).all()
        for eud in euds:
            results['euds'].append(eud[0].to_json())

        markers = db.session.execute(
            db.session.query(Marker).join(CoT).filter(CoT.stale >= datetime.datetime.now(datetime.timezone.utc))).all()
        for marker in markers:
            results['markers'].append(marker[0].to_json())

        rb_lines = db.session.execute(
            db.session.query(RBLine).join(CoT).filter(CoT.stale >= datetime.datetime.now(datetime.timezone.utc))).all()
        for rb_line in rb_lines:
            results['rb_lines'].append(rb_line[0].to_json())

        casevacs = db.session.execute(
            db.session.query(CasEvac).join(CoT).filter(CoT.stale >= datetime.datetime.now(datetime.timezone.utc))).all()
        for casevac in casevacs:
            results['casevacs'].append(casevac[0].to_json())

    except BaseException as e:
        logger.error(traceback.format_exc())
        return jsonify({'success': False, 'error': str(e)}), 500

    return jsonify(results)


@api_blueprint.route('/api/icon')
@auth_required()
def get_icon():
    """Query map icons. All parameters are optional.

    :param filename:
    :param groupName:
    :param type2525b:
    """
    query = db.session.query(Icon)
    query = search(query, Icon, 'filename')
    query = search(query, Icon, 'groupName')
    query = search(query, Icon, 'type2525b')

    return paginate(query)


@api_blueprint.route('/api/itak_qr_string')
@auth_required()
def get_settings():
    """The iTAK QR string in the following format:

    ``OpenTAKServer_SERVER-ADDRESS,SERVER-ADDRESS,8089,SSL``
    """
    url = urlparse(request.url_root).hostname
    return "OpenTAKServer_{},{},{},SSL".format(url, url, app.config.get("OTS_SSL_STREAMING_PORT"))



import datetime
import json
import traceback
import uuid

import bleach
import sqlalchemy.exc
from flask import Blueprint, request, jsonify, current_app as app
from flask_babel import gettext
from flask_security import auth_required, current_user, roles_required, hash_password, verify_password
from xml.etree.ElementTree import tostring
import pika
from sqlalchemy import or_

from opentakserver.blueprints.marti_api.mission_marti_api import invite, generate_mission_delete_cot, generate_new_mission_cot, generate_invitation_cot
from opentakserver.extensions import db, logger
from opentakserver.blueprints.ots_api.api import search, paginate
from opentakserver.models.CoT import CoT
from opentakserver.models.EUD import EUD
from opentakserver.models.Group import Group
from opentakserver.models.GroupMission import GroupMission
from opentakserver.models.GroupUser import GroupUser
from opentakserver.models.Mission import Mission
from opentakserver.models.MissionChange import MissionChange
from opentakserver.models.MissionContentMission import MissionContentMission
from opentakserver.models.MissionInvitation import MissionInvitation
from opentakserver.models.MissionLogEntry import MissionLogEntry
from opentakserver.models.MissionRole import MissionRole
from opentakserver.models.MissionUID import MissionUID

data_sync_api = Blueprint("data_sync_api", __name__)


@data_sync_api.route('/api/missions')
@auth_required()
def get_missions():
    query: db.Query = db.session.query(Mission)
    query = search(query, Mission, 'name')
    query = search(query, Mission, 'guid')
    query = search(query, Mission, 'tool')

    # Only show users missions that belong to the same groups they belong to
    if not current_user.has_role("administrator"):
        group_filters = []
        groups = db.session.execute(db.session.query(GroupUser).filter_by(user_id=current_user.id, direction=Group.IN)).scalars()
        for group in groups:
            group_filters.append(GroupMission.group_id == group.group_id)
        if group_filters:
            query = query.outerjoin(GroupMission).where(or_(*group_filters))

    return paginate(query)


@data_sync_api.route('/api/missions', methods=['PUT', 'POST'])
@auth_required()
def create_edit_mission():
    mission_name = request.json.get('name')
    creator_uid = request.json.get('creator_uid')
    if not mission_name or not creator_uid:
        return jsonify({'success': False, 'error': 'Please provide a mission name and creator UID'}), 400

    mission_name = bleach.clean(mission_name)
    creator_uid = bleach.clean(creator_uid)

    mission = db.session.execute(db.session.query(Mission).filter_by(name=mission_name)).first()

    # Creates a new mission
    if not mission:
        eud = db.session.execute(db.session.query(EUD).filter_by(uid=creator_uid)).first()
        if not eud:
            return jsonify({'success': False, 'error': f"Invalid UID: {creator_uid}"}), 400

        groups = None
        mission_groups = []

        mission = Mission()
        mission.create_time = datetime.datetime.now(datetime.timezone.utc)
        mission.guid = str(uuid.uuid4())
        mission.creator_uid = creator_uid

        for key in request.json.keys():
            if key == 'password' and request.json.get('password'):
                mission.password = hash_password(request.json.get('password'))
            elif key == 'groups':
                group_ids = request.json.get('groups')
                groups = db.session.execute(db.session.query(GroupUser).filter_by(user_id=current_user.id, enabled=True, direction=Group.IN)).scalars()

                # Make sure the user is a member of the IN group that they want the mission to be associated with. Also allow
                # administrators to associate a mission with any group
                for group_id in group_ids:
                    user_in_group = False

                    if not current_user.has_role("administrator"):
                        for group in groups:
                            if group.id == group_id:
                                user_in_group = True
                                break

                    if not user_in_group and not current_user.has_role("administrator"):
                        group = db.session.execute(db.session.query(Group).filter_by(id=group_id)).first()
                        if group:
                            return jsonify({"success": False, "error": f"User is not a member of {group[0].name}"}), 403
                        else:
                            return jsonify({"success": False, "error": f"Invalid group ID: {group_id}"}), 400

                    group_id = int(group_id)
                    group = db.session.execute(db.session.query(Group).filter_by(id=group_id)).first()
                    if not group:
                        continue

                    group_mission = GroupMission()
                    group_mission.mission_name = mission_name
                    group_mission.group_id = group_id
                    mission_groups.append(group_mission)

            elif hasattr(mission, key):
                setattr(mission, key, request.json[key])
            else:
                return jsonify({'success': False, 'error': gettext("Invalid property: %(key)s", key=key)}), 400

        mission.password_protected = (mission.password != '' and mission.password is not None)

        role = MissionRole()
        role.clientUid = creator_uid
        role.username = current_user.username
        role.createTime = datetime.datetime.now(datetime.timezone.utc)
        role.role_type = MissionRole.MISSION_OWNER
        role.mission_name = mission_name

        invitation = MissionInvitation()
        invitation.mission_name = mission_name
        invitation.client_uid = creator_uid
        invitation.creator_uid = creator_uid
        invitation.role = MissionRole.MISSION_OWNER

        try:
            db.session.add(mission)
            db.session.add(role)
            db.session.add(invitation)
            db.session.commit()

            for group in mission_groups:
                db.session.add(group)
            db.session.commit()
        except sqlalchemy.exc.IntegrityError as e:
            logger.error(f"Failed to add mission: {e}")
            logger.debug(mission.serialize())
            return jsonify({'success': False, 'error': gettext("Failed to add mission: %(e)s", e=str(e))}), 400

        rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
        rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
        rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
        channel = rabbit_connection.channel()

        for group in groups:
            logger.error(f"Publishing to {group.group.name}.{group.direction}")
            channel.basic_publish(exchange="groups", routing_key=f"{group.group.name}.{group.direction}", body=json.dumps(
                {'uid': app.config.get("OTS_NODE_ID"), 'cot': tostring(generate_new_mission_cot(mission)).decode('utf-8')}))

        channel.basic_publish(exchange="dms", routing_key=creator_uid,
                              body=json.dumps({'uid': creator_uid,
                                               'cot': tostring(generate_invitation_cot(mission, creator_uid)).decode('utf-8')}))
        channel.close()

        return jsonify({'success': True})

    # Update an existing mission

    # Checks if current user is the mission creator
    mission = mission[0]
    is_user_mission_creator = False
    for eud in current_user.euds:
        if eud.uid == mission.creator_uid:
            is_user_mission_creator = True
            break

    # Only allows admins and the mission creator to change existing missions
    if not current_user.has_role('administrator') and not is_user_mission_creator:
        return jsonify({'success': False, 'error': gettext(u'Only an admin or the mission creator can change this mission')}), 403

    for key in request.json:
        if key == 'groups':
            db.session.execute(sqlalchemy.delete(GroupMission).filter_by(mission_name=mission_name))
            db.session.commit()

            for group_id in request.json.get('groups'):
                group_id = int(group_id)
                group = db.session.execute(db.session.query(Group).filter_by(id=group_id)).first()
                if not group:
                    continue

                group_mission = GroupMission()
                group_mission.mission_name = mission_name
                group_mission.group_id = group_id
                db.session.add(group_mission)
            db.session.commit()
        elif hasattr(mission, key):
            setattr(mission, key, request.json.get(key))
        else:
            return jsonify({'success': False, 'error': gettext(u"Invalid property: %(key)s")}), 400

    db.session.execute(sqlalchemy.update(Mission).filter(Mission.name == mission_name).values(**mission.serialize()))
    db.session.commit()

    return jsonify({'success': True})


@data_sync_api.route('/api/missions', methods=['DELETE'])
@roles_required("administrator")
def delete_mission():
    mission_name = request.args.get('name')
    if not mission_name:
        return jsonify({'success': False, 'error': gettext(u'Please specify a mission name')}), 404

    mission_name = bleach.clean(mission_name)

    mission = db.session.execute(db.session.query(Mission).filter_by(name=mission_name)).first()
    if not mission:
        return jsonify({'success': False, 'error': gettext(u"Mission %(mission_name)s not found", mission_name=mission_name)}), 404
    mission = mission[0]

    db.session.execute(sqlalchemy.delete(GroupMission).where(GroupMission.mission_name == mission_name))
    db.session.execute(sqlalchemy.delete(MissionContentMission).where(MissionContentMission.mission_name == mission_name))
    db.session.execute(sqlalchemy.delete(MissionChange).where(MissionChange.mission_name == mission_name))
    db.session.execute(sqlalchemy.delete(CoT).where(CoT.mission_name == mission_name))
    db.session.execute(sqlalchemy.delete(MissionInvitation).where(MissionInvitation.mission_name == mission_name))
    db.session.execute(sqlalchemy.delete(MissionLogEntry).where(MissionLogEntry.mission_name == mission_name))
    db.session.execute(sqlalchemy.delete(MissionRole).where(MissionRole.mission_name == mission_name))
    db.session.execute(sqlalchemy.delete(MissionUID).where(MissionUID.mission_name == mission_name))
    db.session.execute(sqlalchemy.delete(Mission).where(Mission.name == mission_name))
    db.session.commit()

    rabbit_credentials = pika.PlainCredentials(app.config.get("OTS_RABBITMQ_USERNAME"), app.config.get("OTS_RABBITMQ_PASSWORD"))
    rabbit_host = app.config.get("OTS_RABBITMQ_SERVER_ADDRESS")
    rabbit_connection = pika.BlockingConnection(pika.ConnectionParameters(host=rabbit_host, credentials=rabbit_credentials))
    channel = rabbit_connection.channel()
    channel.basic_publish(exchange="missions", routing_key="missions",
                          body=json.dumps({'uid': app.config.get("OTS_NODE_ID"),
                                           'cot': tostring(generate_mission_delete_cot(mission)).decode('utf-8')}))
    channel.close()
    rabbit_connection.close()

    return jsonify({'success': True})


@data_sync_api.route('/api/missions/invite', methods=['POST'])
@auth_required()
def invite_eud():
    mission_name = request.json['mission_name']
    eud_uid = request.json['uid']
    mission = db.session.execute(db.session.query(Mission).filter_by(name=mission_name)).first()
    if not mission:
        return jsonify({'success': False, 'error': gettext(u"Mission not found: %(mission_name)s", mission_name=mission_name)}), 404
    mission = mission[0]

    # If the user isn't an admin an the mission is password protected, verify the password
    if mission.password_protected and not current_user.has_role("administrator") and not request.json.get('password'):
        return jsonify({'success': False, 'error': gettext(u"Please provide the mission password")}), 403

    elif (mission.password_protected and not current_user.has_role('administrator')
          and not verify_password(request.json.get('password'), mission.password)):
        return jsonify({'success': False, 'error': gettext(u'Invalid password')}), 401

    return invite(mission_name, 'clientuid', eud_uid)



"""Changed device_profile table

Revision ID: 34dc96ee805b
Revises: 414b87c32cc2
Create Date: 2024-09-12 04:42:15.909042

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '34dc96ee805b'
down_revision = '414b87c32cc2'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('device_profiles', schema=None) as batch_op:
        batch_op.add_column(sa.Column('id', sa.Integer(), nullable=False))
        batch_op.alter_column('preference_key',
               existing_type=sa.String(length=255),
               nullable=False)
        batch_op.alter_column('preference_value',
               existing_type=sa.String(length=255),
               nullable=True)
        batch_op.alter_column('value_class',
               existing_type=sa.String(length=255),
               nullable=True)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('device_profiles', schema=None) as batch_op:
        batch_op.alter_column('value_class',
               existing_type=sa.String(length=255),
               nullable=False)
        batch_op.alter_column('preference_value',
               existing_type=sa.String(length=255),
               nullable=False)
        batch_op.alter_column('preference_key',
               existing_type=sa.String(length=255),
               nullable=False)
        batch_op.drop_column('id')

    # ### end Alembic commands ###



import datetime
from dataclasses import dataclass

from opentakserver.extensions import db
from opentakserver.functions import iso8601_string_from_datetime
from sqlalchemy import Integer, String, Boolean, DateTime, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship


@dataclass
class MissionRole(db.Model):
    __tablename__ = "mission_roles"

    MISSION_MANAGE_FEEDS = "MISSION_MANAGE_FEEDS"
    MISSION_SET_PASSWORD = "MISSION_SET_PASSWORD"
    MISSION_WRITE = "MISSION_WRITE"
    MISSION_MANAGE_LAYERS = "MISSION_MANAGE_LAYERS"
    MISSION_UPDATE_GROUPS = "MISSION_UPDATE_GROUPS"
    MISSION_DELETE = "MISSION_DELETE"
    MISSION_SET_ROLE = "MISSION_SET_ROLE"
    MISSION_READ = "MISSION_READ"
    MISSION_OWNER = "MISSION_OWNER"
    MISSION_SUBSCRIBER = "MISSION_SUBSCRIBER"
    MISSION_READ_ONLY = "MISSION_READ_ONLY"

    OWNER_ROLE = {'type': MISSION_OWNER, 'permissions': []}
    OWNER_ROLE['permissions'].append(MISSION_MANAGE_FEEDS)
    OWNER_ROLE['permissions'].append(MISSION_SET_PASSWORD)
    OWNER_ROLE['permissions'].append(MISSION_WRITE)
    OWNER_ROLE['permissions'].append(MISSION_MANAGE_LAYERS)
    OWNER_ROLE['permissions'].append(MISSION_UPDATE_GROUPS)
    OWNER_ROLE['permissions'].append(MISSION_DELETE)
    OWNER_ROLE['permissions'].append(MISSION_SET_ROLE)
    OWNER_ROLE['permissions'].append(MISSION_READ)

    SUBSCRIBER_ROLE = {'type': MISSION_SUBSCRIBER, 'permissions': []}
    SUBSCRIBER_ROLE['permissions'].append(MISSION_READ)
    SUBSCRIBER_ROLE['permissions'].append(MISSION_WRITE)

    READ_ONLY_ROLE = {'type': MISSION_READ_ONLY, 'permissions': []}
    READ_ONLY_ROLE['permissions'].append(MISSION_READ)

    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    clientUid: Mapped[str] = mapped_column(String(255))
    username: Mapped[str] = mapped_column(String(255))
    createTime: Mapped[datetime] = mapped_column(DateTime)
    role_type: Mapped[str] = mapped_column(String(255))
    mission_name: Mapped[str] = mapped_column(String(255), ForeignKey("missions.name"))
    mission = relationship("Mission", back_populates="roles", uselist=False)

    def serialize(self):
        return {
            'clientUid': self.clientUid,
            'username': self.username,
            'createTime': self.createTime,
            'role_type': self.role_type
        }

    def to_json(self):
        json = {
            'clientUid': self.clientUid,
            'username': self.username,
            'createTime': iso8601_string_from_datetime(self.createTime),
            'role': {
                'type': self.role_type,
                'permissions': []
            }
        }

        if self.role_type == self.MISSION_OWNER:
            json['role'] = self.OWNER_ROLE

        elif self.role_type == self.MISSION_SUBSCRIBER:
            json['role'] = self.SUBSCRIBER_ROLE

        else:
            json['role'] = self.READ_ONLY_ROLE

        return json



from dataclasses import dataclass

from sqlalchemy import String, Integer, ForeignKey, Boolean
from sqlalchemy.orm import Mapped, mapped_column, relationship

from opentakserver.extensions import db


@dataclass
class GroupMission(db.Model):
    __tablename__ = "groups_missions"

    mission_name: Mapped[String] = mapped_column(String, ForeignKey("missions.name"), primary_key=True)
    group_id: Mapped[Integer] = mapped_column(Integer, ForeignKey("groups.id"), primary_key=True)
    mission = relationship("Mission", cascade="all, delete", viewonly=True)
    group = relationship("Group", cascade="all, delete", viewonly=True)

    def serialize(self):
        return {
            "mission_name": self.mission_name,
            "group_id": self.group_id,
            "mission": self.mission.serialize(),
            "group": self.group.serialize()
        }



"""Updated groups_users table

Revision ID: ab719ee9b322
Revises: e8bd877719e2
Create Date: 2025-10-28 19:10:38.181780

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ab719ee9b322'
down_revision = 'e8bd877719e2'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('groups', schema=None) as batch_op:
        batch_op.create_unique_constraint("groups_name_key", ['name'])

    with op.batch_alter_table('groups_users', schema=None) as batch_op:
        batch_op.drop_column('id')
        batch_op.create_primary_key("groups_users_pkey", ["user_id", "group_id", "direction"])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('groups_users', schema=None) as batch_op:
        batch_op.add_column(sa.Column('id', sa.INTEGER(), autoincrement=True, nullable=False))

    with op.batch_alter_table('groups', schema=None) as batch_op:
        batch_op.drop_constraint("groups_name_key", type_='unique')

    # ### end Alembic commands ###



import datetime
import os
import time
import traceback
from urllib.parse import urlparse

import bleach
import jwt
from flask import jsonify, request, current_app as app, Blueprint
from flask_babel import gettext
from flask_ldap3_login import AuthenticationResponseStatus
from flask_login import current_user
from flask_security import auth_required, verify_password
from sqlalchemy import delete

from opentakserver.extensions import db, logger, ldap_manager
from opentakserver.models.Token import Token
from opentakserver.models.user import User

token_api_blueprint = Blueprint('token_api_blueprint', __name__)


@token_api_blueprint.route("/oauth/token", methods=['GET', 'POST'])
def cloudtak_oauth_token():
    """
    Provides an OAuth token for TAKX and CloudTAK

    :param username:
    :param password:

    :return: jwt
    """

    username = bleach.clean(request.args.get("username"))
    password = bleach.clean(request.args.get("password"))

    if app.config.get("OTS_ENABLE_LDAP"):
        result = ldap_manager.authenticate(username, password)

        if result.status == AuthenticationResponseStatus.success:
            # Keep this import here to avoid a circular import when OTS is started
            from opentakserver.blueprints.ots_api.ldap_api import save_user

            save_user(result.user_dn, result.user_id, result.user_info, result.user_groups)

        else:
            return jsonify({'success': False, 'error': 'Invalid username or password'}), 400

    else:
        user = app.security.datastore.find_user(username=username)
        if not user or not verify_password(password, user.password):
            return jsonify({'success': False, 'error': 'Invalid username or password'}), 400

    with open(os.path.join(app.config.get("OTS_CA_FOLDER"), "certs", "opentakserver", "opentakserver.nopass.key"),
              "rb") as key:
        token = jwt.encode({
            "exp": datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(days=365),
            "nbf": datetime.datetime.now(datetime.timezone.utc),
            "iss": "OpenTAKServer",
            "aud": "OpenTAKServer",
            "iat": datetime.datetime.now(datetime.timezone.utc),
            "sub": user.username
        }, key.read(), algorithm="RS256")

        return jsonify({"access_token": token, "token_type": "Bearer", "expires_in": 365 * 24 * 60 * 60})


@token_api_blueprint.route("/api/atak_qr_string", methods=['POST'])
@auth_required()
def new_atak_qr_string():
    """
    Generates a QR string for ATAK certificate enrollment. ATAK certificate enrollment via QR code
    only works if your server has a Let's Encrypt certificate. Params are sent as JSON.

    :param username:
    :param exp: The expiration time in unix epoch seconds i.e.1764260510
    :param nbf: Not Before, the token will not be valid until this date in unix epoch seconds.
    :param max: The maximum number of uses for this token.

    :return: String in the format of tak://com.atakmap.app/enroll?host=server_address.com&username=your_username&token=jwt_token
    """
    try:
        username = request.json.get("username") or current_user.username
        if username != current_user.username and not current_user.has_role("administrator"):
            return jsonify({'success': False, 'error': 'Cannot generate QR for another user'}), 401

        else:
            if username != current_user.username and current_user.has_role("administrator"):
                user = db.session.query(User).filter_by(username=username).first()
                if not user:
                    return jsonify({'success': False, 'error': f"No such user: {username}"}), 404

            token = db.session.execute(db.session.query(Token).filter_by(username=username)).first()
            if token:
                token = token[0]
            else:
                token = Token()
                token.creation = int(time.time())

            expiration = int(request.json.get("exp")) if request.json.get("exp") else None
            # PyJWT uses timestamps in seconds to check expiration and not_before
            # This makes sure these timestamps are in seconds and not milliseconds
            # TODO: Fix this before January 19, 2038 03:14:07Z
            if expiration and expiration > 2147483647:
                token.expiration = expiration / 1000
            elif expiration:
                token.expiration = expiration

            not_before = int(request.json.get("nbf")) if request.json.get("nbf") else None
            if not_before and not_before > 2147483647:
                token.not_before = not_before / 1000
            elif not_before:
                token.not_before = not_before

            if request.json.get("max"):
                try:
                    max_uses = int(request.json.get("max"))
                    if max_uses > 0:
                        token.max_uses = max_uses
                except ValueError:
                    pass

            token.username = username
            token.disabled = request.json.get("disabled") if "disabled" in request.json.keys() else None
            token.hash_token()

            db.session.add(token)
            db.session.commit()

            response = token.to_json()
            response["success"] = True
            response["disabled"] = token.disabled
            response["qr_string"] = f"tak://com.atakmap.app/enroll?host={urlparse(request.url_root).hostname}&username={username}&token={token.generate_token()}"
            return jsonify(response)

    except BaseException as e:
        logger.error(f"Failed to create token: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({'success': False, 'error': str(e)}), 400


@token_api_blueprint.route("/api/atak_qr_string", methods=['GET'])
@auth_required()
def get_atak_qr_strings():
    """
    Returns an existing ATAK QR string

    :return: String in the format of tak://com.atakmap.app/enroll?host=server_address.com&username=your_username&token=jwt_token
    """

    query = db.session.query(Token)

    if current_user.has_role("administrator") and request.args.get("username"):
        username = request.args.get("username")
    else:
        username = current_user.username

    query = query.filter_by(username=username)

    token = db.session.execute(query).first()
    if token:
        response = token[0].to_json()
        response["success"] = True
        response["disabled"] = token[0].disabled
        response["total_uses"] = token[0].total_uses
        response["qr_string"] = f"tak://com.atakmap.app/enroll?host={urlparse(request.url_root).hostname}&username={token[0].username}&token={token[0].generate_token()}"
        return jsonify(response)
    else:
        return jsonify({'success': False, 'error': gettext(u"No token found for %s(username)s", username=username)}), 404


@token_api_blueprint.route("/api/atak_qr_string", methods=["DELETE"])
@auth_required()
def delete_token():
    """ Deletes a token

    :parameter username:

    :return: 200 on success
    """
    try:
        if current_user.has_role("administrator") and request.args.get("username"):
            username = request.args.get("username")
        else:
            username = current_user.username

        db.session.execute(delete(Token).where(Token.username == username))
        db.session.commit()

        return jsonify({"success": True})
    except BaseException as e:
        logger.error(f"Failed to delete token: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({"success": False, "error": gettext(u"Failed to delete token: %(e)s", e=str(e))}), 500



"""Added plugins table

Revision ID: f310052349cb
Revises: c9fa07339363
Create Date: 2025-04-24 02:42:00.205913

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f310052349cb'
down_revision = 'c9fa07339363'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('plugins',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('name', sa.String(length=255), nullable=False),
    sa.Column('enabled', sa.Boolean(), nullable=False),
    sa.Column('author', sa.String(length=255), nullable=True),
    sa.Column('version', sa.String(length=255), nullable=True),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('name')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('plugins')
    # ### end Alembic commands ###



from flask_wtf import FlaskForm
from wtforms import StringField, BooleanField
from wtforms.validators import Optional

from opentakserver.functions import false_values


class DeviceProfileForm(FlaskForm):
    preference_key = StringField(validators=[Optional()])
    preference_value = StringField(validators=[Optional()])
    value_class = StringField(validators=[Optional()])
    enrollment = BooleanField(false_values=false_values)
    connection = BooleanField(false_values=false_values)
    tool = StringField(validators=[Optional()])
    active = BooleanField(false_values=false_values)



"""Added tokens table

Revision ID: d68964dab66b
Revises: b0c97a4ce0df
Create Date: 2025-05-07 02:26:29.670335

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'd68964dab66b'
down_revision = 'b0c97a4ce0df'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('tokens',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('username', sa.String(length=255), nullable=True),
                    sa.Column('max_uses', sa.Integer(), nullable=True),
                    sa.Column('total_uses', sa.Integer(), nullable=True),
                    sa.Column('creation', sa.BigInteger(), nullable=False),
                    sa.Column('not_before', sa.BigInteger(), nullable=True),
                    sa.Column('expiration', sa.BigInteger(), nullable=True),
                    sa.Column('disabled', sa.Boolean(), nullable=True),
                    sa.Column('token_hash', sa.String(length=255), nullable=False),
                    sa.ForeignKeyConstraint(['username'], ['user.username'], ),
                    sa.PrimaryKeyConstraint('id'),
                    sa.UniqueConstraint('token_hash'),
                    sa.UniqueConstraint('username')
                    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    op.drop_table('tokens')
    # ### end Alembic commands ###



import glob
import os
import traceback
from urllib.parse import urlparse, unquote

from OpenSSL import crypto
from OpenSSL.crypto import X509
from flask import request, Blueprint, current_app as app, jsonify, send_from_directory
from flask_babel import gettext
from flask_security import current_user
from simplekml import Kml, GxTrack, IconStyle, Icon, Style, GxMultiTrack, Document

from opentakserver.extensions import logger, db
from opentakserver.functions import iso8601_string_from_datetime, datetime_from_iso8601_string
from opentakserver import __version__ as version
from opentakserver.models.EUD import EUD
from opentakserver.models.Point import Point

marti_api = Blueprint('marti_api', __name__)


# Verifies the client cert forwarded by nginx in the X-Ssl-Cert header
# Returns the parsed cert if valid, otherwise returns False
def verify_client_cert() -> X509 | bool:
    cert_header = app.config.get("OTS_SSL_CERT_HEADER")
    if cert_header not in request.headers:
        return False

    cert = unquote(request.headers.get(cert_header))
    cert = crypto.load_certificate(crypto.FILETYPE_PEM, cert)
    with open(os.path.join(app.config.get("OTS_CA_FOLDER"), "ca.pem"), 'rb') as f:
        ca_cert = crypto.load_certificate(crypto.FILETYPE_PEM, f.read())

    store = crypto.X509Store()
    store.add_cert(ca_cert)
    ctx = crypto.X509StoreContext(store, cert)

    try:
        ctx.verify_certificate()
        return cert
    except crypto.X509StoreContextError:
        return False


@marti_api.route('/Marti/api/clientEndPoints', methods=['GET'])
def client_end_points():
    # TODO: Add group support ?group=__ANON__
    euds = db.session.execute(db.select(EUD)).scalars()
    return_value = {'version': 3, "type": "com.bbn.marti.remote.ClientEndpoint", 'data': [],
                    'nodeId': app.config.get("OTS_NODE_ID")}
    for eud in euds:
        if not eud.callsign or not eud.last_event_time:
            continue

        return_value['data'].append({
            'callsign': eud.callsign,
            'uid': eud.uid,
            'username': current_user.username if current_user.is_authenticated else 'anonymous',
            'lastEventTime': iso8601_string_from_datetime(eud.last_event_time),
            'lastStatus': eud.last_status
        })

    return return_value, 200, {'Content-Type': 'application/json'}


@marti_api.route('/Marti/api/version/config', methods=['GET'])
def marti_config():
    url = urlparse(request.url_root)

    return {"version": "3", "type": "ServerConfig",
            "data": {"version": version, "api": "3", "hostname": url.hostname},
            "nodeId": app.config.get("OTS_NODE_ID")}, 200, {'Content-Type': 'application/json'}


@marti_api.route('/Marti/ExportMissionKML')
def atak_track_history():
    try:
        start_time = request.args.get('startTime')
        end_time = request.args.get('endTime')
        uid = request.args.get('uid')
        file_format = request.args.get('format')
        # Not sure what these three are supposed to do
        multitrack_threshold = request.args.get('multiTrackThreshold')
        extended_data = request.args.get('extendedData')
        optimize_export = request.args.get('optimizeExport')

        kml = Kml()
        doc: Document = kml.newdocument(name=uid)

        eud = db.session.execute(db.session.query(EUD).filter_by(uid=uid)).first()
        if not eud:
            return jsonify({'success': False, 'error': gettext("No such UID: %(uid)s", uid=uid)}), 400
        eud: EUD = eud[0]

        icon = Icon(href=f"files/team_{eud.team.name.lower().replace(' ', '')}.png")
        icon_style = IconStyle(scale=1.0, heading=0.0, icon=icon)
        style = Style(iconstyle=icon_style)
        doc.styles.append(style)
        kml.addfile(os.path.join(os.path.dirname(os.path.realpath(__file__)), "icons", f"team_{eud.team.name.lower().replace(' ', '')}.png"))

        query = db.session.query(Point)

        if uid:
            query = query.filter(Point.device_uid == uid)

        if start_time:
            query = query.filter(Point.timestamp >= datetime_from_iso8601_string(start_time))

        if end_time:
            query = query.filter(Point.timestamp <= datetime_from_iso8601_string(end_time))

        points = db.session.execute(query)
        timestamps = []
        coords = []
        for point in points:
            point = point[0]
            timestamps.append(iso8601_string_from_datetime(point.timestamp))
            coords.append((point.longitude, point.latitude))

        multitrack: GxMultiTrack = doc.newgxmultitrack(gxinterpolate=0)
        multitrack.style = style
        multitrack.name = eud.callsign

        track: GxTrack = multitrack.newgxtrack(name=uid, gxaltitudemode="clampToGround")
        track.newwhen(timestamps)
        track.newgxcoord(coords)

        if file_format == "kmz":
            kml.savekmz(os.path.join(app.config.get("UPLOAD_FOLDER"), f"{uid}.kmz"))
            return send_from_directory(app.config.get("UPLOAD_FOLDER"), f"{uid}.kmz", as_attachment=True, download_name=f"{uid}.kmz")
        else:
            kml.save(os.path.join(app.config.get("UPLOAD_FOLDER"), f"{uid}.kml"))
            return send_from_directory(app.config.get("UPLOAD_FOLDER"), f"{uid}.kml", as_attachment=True, download_name=f"{uid}.kml")

    except BaseException as e:
        logger.error(f"Failed to generate KML: {e}")
        logger.debug(traceback.format_exc())
        return jsonify({'success': False, 'error': str(e)}), 500



"""Added mission_change_id column to mission_uids table

Revision ID: 4f0173cdb93b
Revises: 00546817d518
Create Date: 2024-10-11 20:52:40.425560

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '4f0173cdb93b'
down_revision = '00546817d518'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('mission_uids', schema=None) as batch_op:
        batch_op.add_column(sa.Column('mission_change_id', sa.Integer(), nullable=True))
        batch_op.create_foreign_key("mission_change", 'mission_changes', ['mission_change_id'], ['id'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('mission_uids', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='foreignkey')
        batch_op.drop_column('mission_change_id')

    # ### end Alembic commands ###



"""Added mission_logs table

Revision ID: bde915ea136e
Revises: d31bff4a15c7
Create Date: 2024-10-10 18:13:29.597437

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'bde915ea136e'
down_revision = 'd31bff4a15c7'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('mission_logs',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('content', sa.String(length=255), nullable=False),
    sa.Column('creator_uid', sa.String(length=255), nullable=False),
    sa.Column('entry_uid', sa.String(length=255), nullable=False),
    sa.Column('mission_name', sa.String(length=255), nullable=False),
    sa.Column('server_time', sa.DateTime(), nullable=False),
    sa.Column('dtg', sa.DateTime(), nullable=False),
    sa.Column('created', sa.DateTime(), nullable=False),
    sa.Column('content_hash', sa.String(length=255), nullable=True),
    sa.Column('keywords', sa.JSON(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    with op.batch_alter_table('mission_content', schema=None) as batch_op:
        batch_op.alter_column('hash',
               existing_type=sa.String(length=255),
               nullable=False)
        #batch_op.drop_index('hash_index')
        batch_op.create_unique_constraint("hash_index", ['hash'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('mission_content', schema=None) as batch_op:
        batch_op.drop_constraint(None, type_='unique')
        batch_op.create_index('hash_index', ['hash'], unique=1)
        batch_op.alter_column('hash',
               existing_type=sa.String(length=255),
               nullable=True)

    op.drop_table('mission_logs')
    # ### end Alembic commands ###



import typing as t

from flask_security import PasswordUtil


# Prevents @ and : characters in passwords. When using basic auth with RTSP streams these characters cause invalid URLs
class PasswordValidator(PasswordUtil):
    def validate(self, password: str, is_register: bool, **kwargs: t.Any) -> t.Tuple[t.Optional[t.List], str]:
        if '@' in password or ':' in password:
            return ["Passwords should not include @ or : characters"], password
        return super().validate(password, is_register, **kwargs)
